/* -*- mode: c; c-file-style: "linux"; -*- */
/*
    Copyright (C) 2003 Robert Ham <rah@bash.sh>
    Copyright (C) 2001 Paul Davis

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: udpsync_driver.c,v 1.12 2004/03/27 22:08:55 joq Exp $
*/


#include <math.h>
#include <stdio.h>
#include <memory.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/mman.h>

#include <jack/types.h>
#include <jack/internal.h>
#include <jack/engine.h>
#include <sysdeps/time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>


#include "udpsync_driver.h"

#undef DEBUG_WAKEUP


static jack_nframes_t 
udpsync_driver_wait (udpsync_driver_t *driver, int extra_fd, int *status,
		   float *delayed_usecs)
{
	// TODO:
        // first the message is only a char.
	// we should be transmitting sample count or transport info 
	// in the future.
	char msg;
	socklen_t address_size = sizeof(struct sockaddr_in);

	//jack_time_t starting_time = jack_get_microseconds();
	//jack_time_t processing_time = (driver->last_wait_ust?
	//			       starting_time - driver->last_wait_ust,
	//			       0);

	// ok... we wait for a packet on the socket
	// TODO: 
	// we should be able to run freely if the sync source is not transmitting
	// especially we should be able to compensate for packet loss somehow.
	// but lets try this out first.
	
	recvfrom( driver->sockfd, &msg, sizeof(char), 0, &driver->syncsource_address, &address_size );

	driver->srcaddress_valid = 1;

	driver->last_wait_ust = jack_get_microseconds ();
	driver->engine->transport_cycle_start (driver->engine,
					       driver->last_wait_ust);

	/* this driver doesn't work so well if we report a delay */
	*delayed_usecs = 0;		/* lie about it */
	*status = 0;
	return driver->period_size;
}

static inline int
udpsync_driver_run_cycle (udpsync_driver_t *driver)
{
	jack_engine_t *engine = driver->engine;
	int wait_status;
	float delayed_usecs;

	jack_nframes_t nframes = udpsync_driver_wait (driver, -1, &wait_status,
						   &delayed_usecs);
	if (nframes == 0) {
		/* we detected an xrun and restarted: notify
		 * clients about the delay. */
		engine->delay (engine);
		return 0;
	} 

	if (wait_status == 0)
		return engine->run_cycle (engine, nframes, delayed_usecs);

	if (wait_status < 0)
		return -1;
	else
		return 0;
}

static int
udpsync_driver_null_cycle (udpsync_driver_t* driver, jack_nframes_t nframes)
{
	return 0;
}

static int
udpsync_driver_bufsize (udpsync_driver_t* driver, jack_nframes_t nframes)
{
	/* This is a somewhat arbitrary size restriction.  The udpsync
	 * driver doesn't work well with smaller buffer sizes,
	 * apparantly due to usleep() inaccuracy under Linux 2.4.  If
	 * you can get it working with smaller buffers, lower the
	 * limit.  (JOQ) */
	if (nframes < 128)
		return EINVAL;

	driver->period_size = nframes;  
	driver->period_usecs = 
		(jack_time_t) floor ((((float) nframes) / driver->sample_rate)
				     * 1000000.0f);

	/* tell the engine to change its buffer size */
	driver->engine->set_buffer_size (driver->engine, nframes);

	return 0;
}

static int
udpsync_driver_write (udpsync_driver_t* driver, jack_nframes_t nframes)
{
	channel_t chn;
	JSList *node;
	jack_default_audio_sample_t* buf;
	jack_port_t *port;
	int err;

	jack_default_audio_sample_t *packet_buf, *packet_bufX;

	packet_buf = alloca( sizeof(jack_default_audio_sample_t) * 2 * nframes );

	packet_bufX = packet_buf;

	for (chn = 0, node = driver->playback_ports; node;
		node = jack_slist_next (node), chn++) {

	    port = (jack_port_t *) node->data;

	    //if (!jack_port_connected (port)) {
	    //	continue;
	    //  }

	    buf = jack_port_get_buffer (port, nframes);

	    memcpy( packet_bufX, buf, sizeof(jack_default_audio_sample_t) * nframes);
	    packet_bufX += sizeof(jack_default_audio_sample_t) * nframes;
	}

	if( driver->srcaddress_valid )
	    sendto( driver->outsockfd, packet_buf, sizeof(jack_default_audio_sample_t) * 2 * nframes, 0, &driver->syncsource_address, sizeof(struct sockaddr_in) ); 

	return 0;
}


static int
udpsync_driver_attach (udpsync_driver_t *driver)
{
	jack_port_t * port;
	char buf[32];
	unsigned int chn;
	int port_flags;

	driver->engine->set_buffer_size (driver->engine, driver->period_size);
	driver->engine->set_sample_rate (driver->engine, driver->sample_rate);

	port_flags = JackPortIsOutput|JackPortIsPhysical|JackPortIsTerminal;

#if 0
	for (chn = 0; chn < driver->capture_channels; chn++)
	{
		snprintf (buf, sizeof(buf) - 1, "capture_%u", chn+1);

		port = jack_port_register (driver->client, buf,
					   JACK_DEFAULT_AUDIO_TYPE,
					   port_flags, 0);
		if (!port)
		{
			jack_error ("DUMMY: cannot register port for %s", buf);
			break;
		}

		driver->capture_ports =
			jack_slist_append (driver->capture_ports, port);
	}
#endif	
	port_flags = JackPortIsInput|JackPortIsPhysical|JackPortIsTerminal;

	for (chn = 0; chn < driver->playback_channels; chn++)
	{
		snprintf (buf, sizeof(buf) - 1, "playback_%u", chn+1);

		port = jack_port_register (driver->client, buf,
					   JACK_DEFAULT_AUDIO_TYPE,
					   port_flags, 0);

		if (!port)
		{
			jack_error ("DUMMY: cannot register port for %s", buf);
			break;
		}

		driver->playback_ports =
			jack_slist_append (driver->playback_ports, port);
	}

	jack_activate (driver->client);

	return 0;
}

static int
udpsync_driver_detach (udpsync_driver_t *driver)
{
	JSList * node;

	if (driver->engine == 0)
		return 0;
#if 0
	for (node = driver->capture_ports; node; node = jack_slist_next (node))
		jack_port_unregister (driver->client,
				      ((jack_port_t *) node->data));

	jack_slist_free (driver->capture_ports);
	driver->capture_ports = NULL;
#endif
		
	for (node = driver->playback_ports; node; node = jack_slist_next (node))
		jack_port_unregister (driver->client,
				      ((jack_port_t *) node->data));

	jack_slist_free (driver->playback_ports);
	driver->playback_ports = NULL;

	return 0;
}


static void
udpsync_driver_delete (udpsync_driver_t *driver)
{
	jack_driver_nt_finish ((jack_driver_nt_t *) driver);
	free (driver);
}

static jack_driver_t *
udpsync_driver_new (jack_client_t * client,
		  char *name,
		  unsigned int capture_ports,
		  unsigned int playback_ports,
		  jack_nframes_t sample_rate,
		  jack_nframes_t period_size,
		  unsigned int listen_port)
{
	udpsync_driver_t * driver;
	struct sockaddr_in address;

	printf ("creating udpsync driver ... %s|%" PRIu32 "|%" PRIu32
		"|%u|%u|%u\n", name, sample_rate, period_size, listen_port,
		capture_ports, playback_ports);

driver = (udpsync_driver_t *) calloc (1, sizeof (udpsync_driver_t));

	jack_driver_nt_init ((jack_driver_nt_t *) driver);

	driver->write         = (JackDriverReadFunction)       udpsync_driver_write;
	driver->null_cycle    = (JackDriverNullCycleFunction)  udpsync_driver_null_cycle;
	driver->nt_attach     = (JackDriverNTAttachFunction)   udpsync_driver_attach;
	driver->nt_detach     = (JackDriverNTDetachFunction)   udpsync_driver_detach;
	driver->nt_bufsize    = (JackDriverNTBufSizeFunction)  udpsync_driver_bufsize;
	driver->nt_run_cycle  = (JackDriverNTRunCycleFunction) udpsync_driver_run_cycle;

	driver->period_usecs =
		(jack_time_t) floor ((((float) period_size) / sample_rate)
				     * 1000000.0f);
	driver->sample_rate = sample_rate;
	driver->period_size = period_size;
	driver->listen_port   = listen_port;
	driver->last_wait_ust = 0;

	driver->capture_channels  = capture_ports;
	driver->capture_ports     = NULL;
	driver->playback_channels = playback_ports;
	driver->playback_ports    = NULL;

	driver->client = client;
	driver->engine = NULL;

	driver->sockfd = socket( PF_INET, SOCK_DGRAM, 0 );
	if( driver->sockfd == -1 ) {
	    printf( "socket error\n" );
	    return NULL;
	}
	address.sin_family = AF_INET;
	address.sin_port = htons( driver->listen_port );
	address.sin_addr.s_addr = htonl( INADDR_ANY );
	if( bind( driver->sockfd, (struct sockaddr *) &address, sizeof(address) ) < 0 ) {
	    printf( "bind error\n" );
	    return NULL;
	}

	driver->outsockfd = socket( PF_INET, SOCK_DGRAM, 0 );
	if( driver->sockfd == -1 ) {
	    printf( "socket error\n" );
	    return NULL;
	}
	driver->srcaddress_valid = 0;

	return (jack_driver_t *) driver;
}


/* DRIVER "PLUGIN" INTERFACE */

jack_driver_desc_t *
driver_get_descriptor ()
{
	jack_driver_desc_t * desc;
	jack_driver_param_desc_t * params;
	unsigned int i;

	desc = calloc (1, sizeof (jack_driver_desc_t));
	strcpy (desc->name, "udpsync");
	desc->nparams = 5;

	params = calloc (desc->nparams, sizeof (jack_driver_param_desc_t));

	i = 0;
	strcpy (params[i].name, "capture");
	params[i].character  = 'C';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 2U;
	strcpy (params[i].short_desc, "Number of capture ports");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "playback");
	params[i].character  = 'P';
	params[i].type       = JackDriverParamUInt;
	params[1].value.ui   = 2U;
	strcpy (params[i].short_desc, "Number of playback ports");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "rate");
	params[i].character  = 'r';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 48000U;
	strcpy (params[i].short_desc, "Sample rate");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "period");
	params[i].character  = 'p';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 1024U;
	strcpy (params[i].short_desc, "Frames per period");
	strcpy (params[i].long_desc, params[i].short_desc);

	i++;
	strcpy (params[i].name, "listenport");
	params[i].character  = 'l';
	params[i].type       = JackDriverParamUInt;
	params[i].value.ui   = 3000U;
	strcpy (params[i].short_desc,
		"The Port we are listening on for sync packets");
	strcpy (params[i].long_desc, params[i].short_desc);

	desc->params = params;

	return desc;
}

const char driver_client_name[] = "udpsync_pcm";

jack_driver_t *
driver_initialize (jack_client_t *client, const JSList * params)
{
	jack_nframes_t sample_rate = 48000;
	jack_nframes_t period_size = 1024;
	unsigned int capture_ports = 2;
	unsigned int playback_ports = 2;
	unsigned int listen_port = 3000;
	const JSList * node;
	const jack_driver_param_t * param;

	for (node = params; node; node = jack_slist_next (node)) {
  	        param = (const jack_driver_param_t *) node->data;

		switch (param->character) {

		case 'C':
		  capture_ports = param->value.ui;
		  break;

		case 'P':
		  playback_ports = param->value.ui;
		  break;

		case 'r':
		  sample_rate = param->value.ui;
		  break;

		case 'p':
		  period_size = param->value.ui;
		  break;

		case 'l':
		  listen_port = param->value.ui;
		  break;
				
		}
	}

//        if (!wait_time_set)
//	  wait_time = (((float)period_size) / ((float)sample_rate)) * 1000000.0;

	return udpsync_driver_new (client, "udpsync_pcm", capture_ports,
				 playback_ports, sample_rate, period_size,
				 listen_port);
}

void
driver_finish (jack_driver_t *driver)
{
	udpsync_driver_delete ((udpsync_driver_t *) driver);
}
